import logging

def isDataEntry(entry):
    """
    Check if entry is a data entry
    """
    dataEntryKeys = ["type","data","labels","parameters"]
    for key in entry.keys():
        if key not in dataEntryKeys:
            return False
    return True

def getDataEntryLabelIndex(dataEntry,label):
    """
    Expected a data entry
    It returns the index of the label
    """

    logger = logging.getLogger("pyUAMMD")

    for i,lbl in enumerate(dataEntry["labels"]):
        if lbl == label:
            return i

    logger.error(f"Required {label} but is not present.")
    raise Exception(f"Required label is not present.")





import os,sys
import uuid

import copy
import itertools

import json

import logging

from deepdiff import DeepDiff

from .utils.common import getDataEntryLabelIndex

from .utils.update.ids import updateDataEntryIds
from .utils.update.ids import updateComponentSetIds

from .utils.update.groups import getGroupsListEntriesFromComponentSet
from .utils.update.groups import updateComponentSetGroupsLists
from .utils.update.groups import appendGroups

from .appending import appendSystem
from .appending import appendGlobal
from .appending import appendStructure
from .appending import appendState
from .appending import appendIntegrator
from .appending import appendSimulationStep
from .appending import appendForceField
from .appending import appendPatchyParticles

class simulation:

    availModes      = ["simulationId","modelId"]

    availGroupTypes = ["Ids","notIds",
                       "Types","notTypes",
                       "ModelIds","notModelIds",
                       "SimIds","notSimIds",
                       "ModelIdsSimIds","notModelIdsSimIds"]

    id_labels = ["id",
                 "id_i","id_j","id_k","id_l"]

    id_list_labels = ["id_list",
                      "idSet_i","idSet_j","idSet_k","idSet_l"]

    type_labels = ["name",
                   "name_i","name_j"]

    def __init__(self,simulationJSON=None,debug=False):

        self.localID = uuid.uuid1().hex

        ########################

        self.logger = logging.getLogger("pyUAMMD")

        if debug:
            self.logger.setLevel(logging.DEBUG) #<----
        else:
            self.logger.setLevel(logging.INFO) #<----

        self.logger.debug("Initializing simulation, ID: {}".format(self.localID))

        ########################

        if simulationJSON is not None:
            self.sim = copy.deepcopy(simulationJSON)
        else:
            self.sim = {}

        ########################

        #Checks

        stateInSim     = ("state" in self.sim)
        topologyInSim  = ("topology" in self.sim)

        if topologyInSim:
            structureInSim = ("structure" in self.sim["topology"])
        else:
            structureInSim = False

        if stateInSim and not structureInSim:
            self.logger.error("Added state but not structure")
            raise Exception("Added state but not structure")

        if stateInSim and structureInSim:
            Nsim = len(self.sim["state"]["data"])
            Nstr = len(self.sim["topology"]["structure"]["data"])

            if Nsim != Nstr:
                self.logger.error("Number of particles in state and structure does not match")
                raise Exception("Number of particles in state and structure does not match")

        ########################

        self.logger.debug("Simulation created")


    def getID(self):
        return self.localID

    def getNumberOfParticles(self):
        try:
            return len(self.sim["topology"]["structure"]["data"])
        except:
            return 0

    def append(self,sim2app,mode: str = "simulationId") -> None:

        sim2appID = sim2app.getID()

        # Check if ID are the same
        if self.localID == sim2appID:
            self.logger.error("Cannot append simulation to itself")
            raise Exception("Cannot append simulation to itself")

        self.logger.debug(f"Appending \"{sim2appID}\" to \"{self.localID}\" with mode \"{mode}\"")
        if mode not in self.availModes:
            self.logger.error(f"Trying append with mode \"{mode}\", but this mode is not available. " \
                             f"Available modes are: {self.availModes}")
            raise Exception(f"Trying append with not available mode \"{mode}\"")

        if mode == "simulationId":
            #Check both simulations have exactly the same entries.
            #In principle this is not compulsory but avoid a large
            #number of errors.

            def get_keys(dictionary):
                result = []
                for key, value in dictionary.items():
                    if type(value) is dict:
                        new_keys = get_keys(value)
                        result.append(key)
                        for innerkey in new_keys:
                            result.append(f'{key}/{innerkey}')
                    else:
                        result.append(key)
                return result

            simEntries     = get_keys(self.sim)
            sim2appEntries = get_keys(sim2app.sim)

            #Remove all simulationStep entries

            simEntries = [entry for entry in simEntries if "simulationStep" not in entry]
            sim2appEntries = [entry for entry in sim2appEntries if "simulationStep" not in entry]

            if len(simEntries) != len(sim2appEntries):
                self.logger.error("Trying to append simulations with different entries while using mode \"simulationId\"")
                raise Exception("Trying to append simulations with different entries")
            else:
                if not (sorted(simEntries) == sorted(sim2appEntries)):
                    #Compute not shared entries
                    self.logger.error("Trying to append simulations with different entries (but same length) while using mode \"simulationId\"")
                    raise Exception("Trying to append simulations with different entries")

        ########################################

        #For avoiding problems patchy particles are appended first
        #We need reference sim structure to be unmodified
        appendPatchyParticles("topology",
                              "structure",
                              "forceField",
                              self,sim2app,mode,
                              self.availGroupTypes,
                              self.id_labels, self.id_list_labels,self.type_labels)

        ###########################################################


        appendSystem("system",self,sim2app,mode)
        appendGlobal("global",self,sim2app,mode)

        idOffset,modeOffset = appendStructure("topology","structure",self,sim2app,mode)

        appendState("topology",
                    "structure",
                    "state",self,sim2app,mode,idOffset,modeOffset)
        appendIntegrator("topology",
                         "structure",
                         "integrator",self,sim2app,mode,idOffset,modeOffset,self.availGroupTypes)

        appendSimulationStep("topology",
                             "structure",
                             "simulationStep",
                             self,sim2app,mode,idOffset,modeOffset,self.availGroupTypes,
                             self.id_labels, self.id_list_labels)
        appendForceField("topology",
                         "structure",
                         "forceField",
                         self,sim2app,mode,idOffset,modeOffset,self.availGroupTypes,
                         self.id_labels, self.id_list_labels,self.type_labels,
                         ignoredEntriesType=["PatchyParticles"])

    def write(self,output:str):
        self.logger.debug("Writting to json...")

        try:
            import JFIO
            self.logger.debug("Writting with JFIO")
            JFIO.write(output,self.sim,formatted=True)
        except:
            import jsbeautifier
            self.logger.warning("Writting with legacy")
            with open(output,"w") as f:
                opts = jsbeautifier.default_options()
                opts.indent_size  = 2
                f.write(jsbeautifier.beautify(json.dumps(self.sim), opts))

    def setValue(self,path,value):
        simTmp = self.sim
        for p in path[:-1]:
            simTmp = simTmp[p]
        simTmp[path[-1]] = value

    #With the following methods simulation behaves like a dict

    def __getitem__(self,key):
        return self.sim[key]

    def __setitem__(self,key,value):
        self.sim[key]=value

    def __delitem__(self, key):
        del self.sim[key]

    def __contains__(self, key):
        return key in self.sim

    def keys(self):
        return list(self.sim.keys())

    def values(self):
        return list(self.sim.values())

    def items(self):
        return list(self.sim.items())

    def pop(self, key, default=None):
        try:
            value = self.sim[key]
            del self.sim[key]
            return value
        except KeyError:
            if default is not None:
                return default
            else:
                raise

    def run(self):
        self.logger.debug("Runing simulation...")

        try:
            from .utils.launcher.UAMMDlauncher import UAMMDlauncher
            UAMMDlauncher.UAMMDlauncher(self.sim)
        except:
            self.logger.error("Something went wrong with UAMMDlauncher")



import os
import logging

class CustomFormatter(logging.Formatter):

    white = "\x1b[37;20m"
    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format     = "%(asctime)s - %(name)s - %(levelname)s - %(message)s "
    formatLine = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

    FORMATS = {
        logging.DEBUG: white + formatLine + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + formatLine + reset,
        logging.CRITICAL: bold_red + formatLine + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt,datefmt='%d/%m/%Y %H:%M:%S')

        return formatter.format(record)

# Set up logger
logger = logging.getLogger("pyUAMMD")
logger.setLevel(logging.INFO)

clog = logging.StreamHandler()

clog.setFormatter(CustomFormatter())
logger.addHandler(clog)

from .simulation import simulation





import copy
import logging

from deepdiff import DeepDiff

def getGroupsListEntriesFromComponentSet(componentSet):
    """
    Returns a list with all the entries which class is "Groups" and subClass "GroupsList"
    """
    groups = []
    for name,entry in componentSet.items():
        if "type" in entry:
            if entry["type"][0] == "Groups":
                if entry["type"][1] == "GroupsList":
                    groups.append(name)

    return groups

def updateComponentSetGroupsLists(mode,componentSet,idOffset,modeOffset,availGroupTypes):
    """
    This function take a component set. It looks for all the groups list entries
    in the component set. Then, if some groups list is found, it updates the
    group definition according to the mode and values of idOffset and modeOffset.

    Returns a copy of the component set with the updated groups.
    """

    logger = logging.getLogger("pyUAMMD")

    updatedComponentSet = copy.deepcopy(componentSet)

    groupsLists = getGroupsListEntriesFromComponentSet(updatedComponentSet)

    #If not, doing nothing
    if len(groupsLists) == 0:
        return updatedComponentSet

    for name in groupsLists:
        logger.debug("Updating groups list \"{}\"".format(name))

        groupDeclarations = updatedComponentSet[name]["data"]

        for groupName,groupType,selection in groupDeclarations:
            if groupType not in availGroupTypes:
                logger.error("Group type \"{}\" is not supported.".format(groupType))
                raise Exception("Group type not supported.")
            else:
                if groupType in ["Ids","notIds"]:
                    for n,_ in enumerate(selection):
                        selection[n] += idOffset
                elif groupType in ["Types","notTypes"]:
                    #Do nothing
                    pass
                elif groupType in ["ModelIds","notModelIds"]:
                    if mode == "modelId":
                        for n,_ in enumerate(selection):
                            selection[n] += modeOffset
                    else:
                        #Do nothing
                        pass
                elif groupType in ["SimIds","notSimIds"]:
                    if mode == "simulationId":
                        for n,_ in enumerate(selection):
                            selection[n] += modeOffset
                    else:
                        #Do nothing
                        pass
                elif groupType in ["ModelIdsSimIds","notModelIdsSimIds"]:
                    if mode == "modelId":
                        for n,_ in enumerate(selection):
                            selection[n][0] += modeOffset
                    elif mode == "simulationId":
                        for n,_ in enumerate(selection):
                            selection[n][1] += modeOffset
                    else:
                        #Do nothing
                        pass

    return updatedComponentSet


def appendGroups(mode,componentRefSet,component2AppendSet,idOffset,modeOffset,availGroupTypes):
    """
    This function takes two component set. It updates all the groupsList entries found
    in the component set to append according to the mode and values of idOffset and modeOffset.
    Then it appends the groupsList entries of the component set to append to the component set
    of reference.

    It does not return anything. Reference component set is updated.

    Rules:
    *No groups list in reference component set:
    -- No groups list in append component set -> Do nothing
    -- Groups list in append component set -> Update all components of the group list type in the set of components to be added and append them to reference
    *Groups list in reference component set:
    -- No groups list in append component set -> Do nothing
    -- Groups list in append component set ->
            1) Update all components of the group list type in the set of components to be added
            2) If the name of some group list type in the set of components to be added
               match some name of some group list type in the reference set of components ->
               1) If both group list types are the same -> Do nothing
               2) If both group list types are different -> Raise an error

    """

    logger = logging.getLogger("pyUAMMD")

    #Check if a entry of class "Groups" and subClass "GroupsList" is present
    groupsListsRef     = getGroupsListEntriesFromComponentSet(componentRefSet)
    groupsLists2Append = getGroupsListEntriesFromComponentSet(component2AppendSet)

    #If not, doing nothing
    if len(groupsListsRef) == 0:
        if len(groupsLists2Append) == 0:
            return
        else:
            updatedComponent2AppendSet = updateComponentSetGroupsLists(mode,component2AppendSet,idOffset,modeOffset,availGroupTypes)
            for name in groupsLists2Append:
                componentRefSet[name] = copy.deepcopy(updatedComponent2AppendSet[name])
    else:
        if len(groupsLists2Append) == 0:
            return
        else:
            updatedComponent2AppendSet = updateComponentSetGroupsLists(mode,component2AppendSet,idOffset,modeOffset,availGroupTypes)

            for name in groupsLists2Append:
                if name in groupsListsRef:

                    refGroupDeclarationNames      = [x[0] for x in componentRefSet[name]["data"]]
                    _2AppendGroupDeclarationNames = [x[0] for x in component2AppendSet[name]["data"]]

                    #Check there are no repeteaded names
                    if len(refGroupDeclarationNames) != len(set(refGroupDeclarationNames)):
                        logger.error("There are repeated group declaration names in the component set of reference.")
                        raise Exception("Repeated group declaration names.")
                    if len(_2AppendGroupDeclarationNames) != len(set(_2AppendGroupDeclarationNames)):
                        logger.error("There are repeated group declaration names in the component set to append.")
                        raise Exception("Repeated group declaration names.")

                    for gName2Append in _2AppendGroupDeclarationNames:
                        if gName2Append in refGroupDeclarationNames:
                            gRef = [gDecl for gDecl in componentRefSet[name]["data"] if gDecl[0] == gName2Append]
                            gApp = [gDecl for gDecl in updatedComponent2AppendSet[name]["data"] if gDecl[0] == gName2Append]

                            equal = not DeepDiff(gRef,gApp,ignore_order=True)
                            if equal:
                                continue
                            else:
                                logger.error("Group \"{}\" already present in \"{}\".".format(gName2Append,name))
                                raise Exception("Group already present.")
                        else:
                            g = [gDecl for gDecl in updatedComponent2AppendSet[name]["data"] if gDecl[0] == gName2Append]

                            componentRefSet[name]["data"].append(g[0])
                else:
                    for name in groupsLists2Append:
                        componentRefSet[name] = copy.deepcopy(updatedComponent2AppendSet[name])



import copy
import logging

from ..utils.common import getDataEntryLabelIndex

def appendGlobal(globalKey,sim,sim2app,mode):

    logger = logging.getLogger("pyUAMMD")

    #########################
    #Appending global data
    #Global data structure:
    #   globalKey: {
    #       "parameters": {"temperature": ..., "box": [...,...,...], ...},
    #       "labels": ["name", "mass", "radius", "charge", ...],
    #       "data": [[...],
    #                [...],
    #                ...]

    # Four cases:
    # 1) global NOT in sim and global NOT in sim2app
    # 2) global NOT in sim and global in sim2app
    # 3) global in sim and global NOT in sim2app
    # 4) global in sim and global in sim2app

    globalInSim     = (globalKey in sim)
    globalInSim2app = (globalKey in sim2app)

    # 1) global NOT in sim and global NOT in sim2app
    if not globalInSim and not globalInSim2app:
        #Do nothing
        pass
    # 2) global NOT in sim and global in sim2app
    elif not globalInSim and globalInSim2app:
        sim[globalKey] = copy.deepcopy(sim2app[globalKey])
    # 3) global in sim and global NOT in sim2app
    elif globalInSim and not globalInSim2app:
        #Do nothing
        pass
    # 4) global in sim and global in sim2app
    elif globalInSim and globalInSim2app:

        #Process parameters

        parametersInSim     = ("parameters" in sim[globalKey])
        parametersInSim2app = ("parameters" in sim2app[globalKey])

        if not parametersInSim and not parametersInSim2app:
            #Do nothing
            pass
        elif not parametersInSim and parametersInSim2app:
            sim[globalKey]["parameters"] = copy.deepcopy(sim2app[globalKey]["parameters"])
        elif parametersInSim and not parametersInSim2app:
            #Do nothing
            pass
        elif parametersInSim and parametersInSim2app:

            #Check if parameters share some keys. If yes, check if they are the same.
            #If parameters do not share some keys or the shared keys are the same, append them.
            for key in sim[globalKey]["parameters"]:
                if key in sim2app[globalKey]["parameters"]:
                    if sim[globalKey]["parameters"][key] != sim2app[globalKey]["parameters"][key]:
                        logger.error("Shared global parameters are not the same")
                        raise Exception("Shared global parameters are not the same")
            sim[globalKey]["parameters"].update(sim2app[globalKey]["parameters"])

        #Process types
        typesInSim     = ("labels" in sim[globalKey])
        typesInSim2app = ("labels" in sim2app[globalKey])

        if not typesInSim and not typesInSim2app:
            #Do nothing
            pass
        elif not typesInSim and typesInSim2app:
            sim[globalKey]["labels"] = copy.deepcopy(sim2app[globalKey]["labels"])

            if "data" not in sim[globalKey]:
                sim[globalKey]["data"] = []

            sim[globalKey]["data"].extend(copy.deepcopy(sim2app[globalKey]["data"]))
        elif typesInSim and not typesInSim2app:
            #Do nothing
            pass
        elif typesInSim and typesInSim2app:

            #Check if global data are the same
            if sim[globalKey]["labels"] != sim2app[globalKey]["labels"]:
                logger.error("Global data labels are not the same")
                raise Exception("Global data labels are not the same")

            #At this point labels and parameters are the same, so we can append data.
            nameIndex     = getDataEntryLabelIndex(sim[globalKey],"name")
            nameIndex2app = getDataEntryLabelIndex(sim2app[globalKey],"name")

            #Iterate over data in sim2app
            for d in sim2app[globalKey]["data"]:
                #Check if name is already in sim. If not, append data.
                #If name is already in sim, check if data are the same. If not, raise error.
                if d[nameIndex2app] not in [x[nameIndex] for x in sim[globalKey]["data"]]:
                    sim[globalKey]["data"].append(d)
                else:
                    for x in sim[globalKey]["data"]:
                        if x[nameIndex] == d[nameIndex2app]:
                            if x != d:
                                logger.error("Shared type name. But type data are not the same")
                                raise Exception("Shared type name. But type data are not the same")

    #Global data appended
    #########################



import copy

import logging

from ..utils.common import getDataEntryLabelIndex

def appendState(topologyKey,
                structureKey,
                stateKey,
                sim,sim2app,mode,idOffset,modeOffset):

    logger = logging.getLogger("pyUAMMD")

    #########################
    #Appending state
    #State structure:
    #   stateKey: {
    #       "labels": ["id","position",...]
    #       "data": [[...],
    #                [...],
    #                ...]

    # Four cases:
    # 1) state NOT in sim and state NOT in sim2app
    # 2) state NOT in sim and state in sim2app
    # 3) state in sim and state NOT in sim2app
    # 4) state in sim and state in sim2app

    stateInSim     = (stateKey in sim)
    stateInSim2app = (stateKey in sim2app)

    structureInSim     = (topologyKey in sim)
    if structureInSim:
        structureInSim     = (structureKey in sim[topologyKey])

    if (stateInSim or stateInSim2app) and not structureInSim:
        logger.warning("Structure must be present in at least one simulation")
        logger.error("Cannot merge state without any structure")
        raise Exception("Cannot merge state without structure")

    # 1) state NOT in sim and state NOT in sim2app
    if not stateInSim and not stateInSim2app:
        #Do nothing
        pass
    # 2) state NOT in sim and state in sim2app
    elif not stateInSim and stateInSim2app:
        sim[stateKey] = copy.deepcopy(sim2app[stateKey])
    # 3) state in sim and state NOT in sim2app
    elif stateInSim and not stateInSim2app:
        #Do nothing
        pass
    # 4) state in sim and state in sim2app
    elif stateInSim and stateInSim2app:
        #At this point, idOffset should be defined from the structure merging

        #Handle labels
        simLabels     = sim[stateKey]["labels"]
        sim2appLabels = sim2app[stateKey]["labels"]

        availLabels = ["position","velocity","direction"]
        availLabelZeros = {"position":[0.0,0.0,0.0],
                           "velocity":[0.0,0.0,0.0],
                           "direction":[0.0,0.0,0.0,1.0]}

        if "id" not in simLabels or "id" not in sim2appLabels:
            logger.error("Cannot merge state without id")
            raise Exception("Cannot merge state without id")

        for label in availLabels:
            labelInSim     = (label in simLabels)
            labelInSim2app = (label in sim2appLabels)

            if not labelInSim and not labelInSim2app:
                #Do nothing
                pass
            elif not labelInSim and labelInSim2app:
                sim[stateKey]["labels"].append(label)
                for d in sim[stateKey]["data"]:
                    d.append(availLabelZeros[label])
            elif labelInSim and not labelInSim2app:
                sim2app[stateKey]["labels"].append(label)
                for d in sim2app[stateKey]["data"]:
                    d.append(availLabelZeros[label])
            elif labelInSim and labelInSim2app:
                #Do nothing
                pass

        #Data merging
        for stateInfo in sim2app[stateKey]["data"]:
            d = []
            for label in sim[stateKey]["labels"]:
                if label == "id":
                    d.append(stateInfo[sim2appLabels.index(label)] + idOffset)
                else:
                    d.append(stateInfo[sim2appLabels.index(label)])
            sim[stateKey]["data"].append(d)

    #State appended
    #########################

